import os
import asyncio
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

from werk24._version import __version__
from werk24.cli.techread import _get_drawing
from werk24.exceptions import RequestTooLargeException
from werk24.gui import event_illustrator
from werk24.gui.measure_table import W24GuiMeasureTable
from werk24.gui.event_feed import W24GuiEventFeed
from werk24.gui.style import ft_headline
from werk24.gui.worker import W24GuiWorker, W24GuiWorkerSignals
from werk24.models.ask import (W24AskCanvasThumbnail, W24AskPageThumbnail,
                               W24AskSectionalThumbnail, W24AskSheetThumbnail,
                               W24AskVariantGDTs, W24AskVariantMeasures)
from werk24.models.measure import W24Measure
from werk24.models.gdt import W24GDT
from werk24.gui.gdt_table import W24GuiGdtTable
from werk24.models.techread import (W24TechreadMessage,
                                    W24TechreadMessageSubtypeError,
                                    W24TechreadMessageSubtypeProgress,
                                    W24TechreadMessageType)
from werk24.techread_client import Hook, LicenseError, W24TechreadClient

try:
    from PyQt5.QtCore import QSize, QState, Qt, QThreadPool
    from PyQt5.QtGui import QIcon
    from PyQt5.QtWidgets import (
        QApplication,
        QCheckBox,
        QFileDialog,
        QFrame,
        QGridLayout,
        QLabel,
        QLayout,
        QMainWindow,
        QMessageBox,
        QPushButton,
        QScrollArea,
        QSizePolicy,
        QSpacerItem,
        QVBoxLayout,
        QWidget)
except ImportError:
    print("Using the W24GUI requires the installation of PyQt5."
          "The installer take care of this, if you call "
          "`pip install werk24[gui]`")
    sys.exit(0)


try:
    # get more meaningful segmentation fault messages
    import faulthandler
    faulthandler.enable()
except ImportError:
    pass

EXTENSIONS_SUPPORTED = ["JPG", "JPEG", "PDF", "PNG", "TIF", "TIFF"]
""" List of all the supported file extensions. """
# TODO: mid-term we want to obtain this from the API ... support will grow

# Set the windows details
try:
    import ctypes
    myappid = f"werk24.techread.gui.{__version__}"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

# This will fail with either an attribute error or an import error
# depending on the operating system we are dealing with
except (AttributeError, ImportError):
    pass


class W24Gui(QMainWindow):
    """ Main GUI Class handling all the interaction
    with the user
    """

    def __init__(self) -> None:
        super().__init__()

        # start the the threadpool and the associated
        # signals to ensure that we can communciate
        # with the workers
        self.threadpool = QThreadPool()
        self.signals = W24GuiWorkerSignals()
        self.signals.result.connect(self._handle_result_signal)
        self.signals.completion.connect(self._handle_completion_signal)

        # make a list of all the asks
        # that the user can request
        self.ask_page_thumbnail = False
        self.ask_sheet_thumbnail = False
        self.ask_canvas_thumbnail = False
        self.ask_sectional_thumbnail = False
        self.ask_variant_measures = False
        self.ask_variant_gdts = False

        # make a store for sectionsals
        self.sectional_thumbnails: Dict[str, Optional[bytes]] = {}

        # and indiate the indicidual parts
        # of the GUI
        self._init_left_grid()
        self._init_right_grid()
        self._init_ui()

        # finally initiate the client
        self.init_client()

    def init_client(self) -> None:
        """ Initiate the TechreadClient from the
        environment variables.

        TODO: if we are not able to do this, we might want to
        TODO give the user a second chance an chose a  license file
        """
        try:
            self.client = W24TechreadClient.make_from_env()

            # Update the Status bar to reflect the current username
            username = self.client.username
            self.statusBar().showMessage(f"Logged in as {username}")

        # If a license error occured, let the user know.
        except LicenseError as exception:
            self._warn(f"LICENSE ERROR: {exception}")
            sys.exit()

    def _init_left_grid(self) -> None:
        """ Initiate the left grid which contains all the ASKS and
        allows the user to start the process
        """

        # STEP 1 FILENAME
        self.lb_input_file = QLabel("Input File")
        self.lb_input_file.setFont(ft_headline)

        # select the file
        self.bt_input_file = QPushButton("Select File and start")
        self.bt_input_file.clicked.connect(self.pick_input_file)

        # STEP 2: ASKS
        self.sp_asks = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.lb_asks = QLabel("Asks")
        self.lb_asks.setFont(ft_headline)

        # ASK PAGE THUMBNAIL
        self.cb_ask_page_thumbnail = QCheckBox('Page Thumbnail', self)
        self.cb_ask_page_thumbnail.stateChanged.connect(
            self._change_ask_page_thumbnail)
        self.cb_ask_page_thumbnail.toggle()

        # ASK SHEET THUMBNAIL
        self.cb_ask_sheet_thumbnail = QCheckBox('Sheet Thumbnail', self)
        self.cb_ask_sheet_thumbnail.stateChanged.connect(
            self._change_ask_sheet_thumbnail)

        # ASK CANVAS THUMBNAIL
        self.cb_ask_canvas_thumbnail = QCheckBox('Canvas Thumbnail', self)
        self.cb_ask_canvas_thumbnail.stateChanged.connect(
            self._change_ask_canvas_thumbnail)

        # ASK SECTIONAL THUMBNAIL
        self.cb_ask_sectional_thumbnail = QCheckBox(
            'Sectional Thumbnail', self)
        self.cb_ask_sectional_thumbnail.stateChanged.connect(
            self._change_ask_sectional_thumbnail)

        # ASK VARIANT MEASURES
        self.cb_ask_variant_measures = QCheckBox(
            'Variant Measures', self)
        self.cb_ask_variant_measures.stateChanged.connect(
            self._change_ask_variant_measures)

        # ASK VARIANT GDTS
        self.cb_ask_variant_gdts = QCheckBox(
            'Variant GD&Ts', self)
        self.cb_ask_variant_gdts.stateChanged.connect(
            self._change_ask_variant_gdts)

        # STEP 2: ASKS
        self.sp_start = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        # Make the left grid with all the checkboxes that allow
        # the user to decide what feedback they want to obtain
        self.left_grid = QVBoxLayout()
        self.left_grid.addWidget(self.lb_asks)
        self.left_grid.addWidget(self.cb_ask_page_thumbnail)
        self.left_grid.addWidget(self.cb_ask_sheet_thumbnail)
        self.left_grid.addWidget(self.cb_ask_canvas_thumbnail)
        self.left_grid.addWidget(self.cb_ask_sectional_thumbnail)
        self.left_grid.addWidget(self.cb_ask_variant_measures)
        self.left_grid.addWidget(self.cb_ask_variant_gdts)
        self.left_grid.addItem(self.sp_asks)
        self.left_grid.addWidget(self.lb_input_file)
        self.left_grid.addWidget(self.bt_input_file)
        self.left_grid.setSizeConstraint(QLayout.SetFixedSize)

        self.left_grid.setAlignment(
            Qt.AlignLeading | Qt.AlignLeft | Qt.AlignTop)
        # Create a frame out f the left grid
        self.left_frame = QFrame()
        self.left_frame.setLayout(self.left_grid)

    def _init_right_grid(self) -> None:
        """ Initiate teh right grid that collects all
        the API responses
        """

        # make the right grid. This will be empty at
        # the start but fill up with data once we get
        # the responses
        self.api_feed = W24GuiEventFeed()

        # turn the right grid into a frame
        self.right_frame = QFrame()
        self.right_frame.resize(550, 550)
        self.right_frame.setLayout(self.api_feed)

        self.right_scroll = QScrollArea()
        self.right_scroll.setWidget(self.right_frame)
        self.right_scroll.setWidgetResizable(True)
        self.api_feed.add_welcome()

    def _init_ui(self) -> None:
        """ Initiate the main part of the GUI
        """

        # Main Grid
        self.main_grid = QGridLayout()
        self.main_grid.setColumnMinimumWidth(10, 0)
        self.main_grid.setAlignment(Qt.AlignTop)
        self.main_grid.addWidget(self.left_frame)
        self.main_grid.addWidget(self.right_scroll)

        self.central_widget = QWidget()
        self.central_widget.setLayout(self.main_grid)

        # Main Interface
        self.setGeometry(300, 200, 850, 450)
        self.setCentralWidget(self.central_widget)
        self.setWindowTitle('Werk24')
        self._set_window_icons()
        self.show()

    def _set_window_icons(self) -> None:
        """ Set the window icons
        """
        path_cur = os.path.dirname(os.path.abspath(__file__))
        path_img = os.path.join(path_cur, "..", "assets", "images")

        app_icon = QIcon()
        for width in [16, 24, 32, 48, 256]:
            icon_path = os.path.join(path_img, f"favicon-{width}x{width}.png")
            app_icon.addFile(icon_path, QSize(width, width))
        self.setWindowIcon(app_icon)

    def _change_ask_page_thumbnail(self, state: QState) -> None:
        self.ask_page_thumbnail = bool(state)

    def _change_ask_sheet_thumbnail(self, state: QState) -> None:
        self.ask_sheet_thumbnail = bool(state)

    def _change_ask_canvas_thumbnail(self, state: QState) -> None:
        self.ask_canvas_thumbnail = bool(state)

    def _change_ask_sectional_thumbnail(self, state: QState) -> None:
        self.ask_sectional_thumbnail = bool(state)

    def _change_ask_variant_measures(self, state: QState) -> None:
        self.ask_variant_measures = bool(state)

    def _change_ask_variant_gdts(self, state: QState) -> None:
        self.ask_variant_gdts = bool(state)

    def pick_input_file(self) -> None:
        """ Let the user chose an input file and
        start the techread process
        """

        # make a list of the supporte file extensions
        # that the user can select
        extensions = " ".join([
            f"*.{ft.lower()}"
            for ft in EXTENSIONS_SUPPORTED])

        # open the file dialog
        # NOTE: the user can only select one file!
        input_file = QFileDialog.getOpenFileName(
            self,
            'Select drawing',
            '',
            f"Technical Drawings ({extensions})")

        # ignore the picked file if it empty (e.g., when the user
        # did not end up picking a file)
        if len(input_file) < 2 or input_file[0] == '':
            return

        # disable the button to make it less likely that they
        # user starts a second parallel request
        self.bt_input_file.setEnabled(False)

        # reset the local stores
        self.sectional_thumbnails = {}

        # clear the feed and give some information that
        # we now have started processing the file in
        # question
        self.api_feed.clear()
        self.api_feed.add_line()
        self.api_feed.add_headline(f"Reading file '{input_file[0]}'")

        # otherwise we want to update the filename and start
        # the techread request
        hooks = self._make_hooks()
        worker = W24GuiWorker(
            self._start_techread_request,
            input_file[0],
            hooks)
        self.threadpool.start(worker)

    def _start_techread_request(
            self,
            input_file: str,
            hooks: List[Hook]
    ) -> None:
        """ Just a small handler to make the code more readable.
        The Client itself works in an asyncronic fashion, which
        somewhat collides with single-task workers.

        Args:
            input_file (str): Path of the input file
            hooks (List[Hook]): List of all the hooks to
                call when data becomdes availabel
        """

        loop = asyncio.new_event_loop()
        loop.run_until_complete(
            self._start_techread_request_async(input_file, hooks))
        try:
            loop.close()
        except RunTimeError:
            pass

    async def _start_techread_request_async(
            self,
            input_file: str,
            hooks: List[Hook]
    ) -> None:
        """ Star the techread request by sending the input file
        to the client. The responses will be handled by the
        associated hooks

        Args:
            input_file (str): Path of the input file
            hooks (List[Hook]): List of all the hooks to
                call when data becomdes availabel
        """

        # then start the session
        async with self.client as session:

            # and make the request
            try:
                drawing_bytes = _get_drawing(input_file)
                if drawing_bytes is None:
                    return

                await session.read_drawing_with_hooks(drawing_bytes, hooks)

            except FileNotFoundError:
                self._warn("File was not found")

            except RequestTooLargeException:
                self._warn(
                    "Request was too large to be processed. " +
                    "Please check the documentation for current limits.")

        # finaly tell the system that the request is done
        self.signals.completion.emit()

    @staticmethod
    def _warn(message: str) -> None:
        """ Create a small warning pop up 1990s-style

        Args:
            message (str): Warning message
        """

        msg = QMessageBox()
        msg.setIcon(QMessageBox.Warning)
        msg.setText(message)
        msg.setWindowTitle("Warning")
        msg.exec_()

    def _handle_completion_signal(
            self
    ) -> None:

        # reactivate the input button
        self.bt_input_file.setEnabled(True)

    @staticmethod
    def _handle_result_signal(
            signal: Tuple[Callable, W24TechreadMessage]
    ) -> None:
        """ Once we obtain a single, we want to collect the
        result and change the state of the GUI. The method
        waits for the signal and then executes the function
        in this thread.

        Args:
            signal (Tuple[Callable, W24TechreadMessage]):
                Signal with Callable and Message
        """
        function, message = signal
        function(message)

    def _make_hooks(self) -> List[Hook]:
        """ Make the hooks depending on the asks that
        the user has selected.

        Returns:
            List[Hook]: List of hooks with associated
                asks and functions
        """

        # tell the api what asks you are interested in,
        # and define what to do when you receive the result
        hooks = []

        config = [
            {
                'condition': self.ask_page_thumbnail,
                'ask': W24AskPageThumbnail(),
                'function': self._receive_ask_page_thumbnail
            },
            {
                'condition': self.ask_sheet_thumbnail,
                'ask': W24AskSheetThumbnail(),
                'function': self._receive_ask_sheet_thumbnail
            },
            {
                'condition': self.ask_canvas_thumbnail,
                'ask': W24AskCanvasThumbnail(),
                'function': self._receive_ask_canvas_thumbnail
            },
            {
                'condition': self.ask_sectional_thumbnail,
                'ask': W24AskSectionalThumbnail(),
                'function': self._receive_ask_sectional_thumbnail
            },
            {
                'condition': self.ask_variant_measures,
                'ask': W24AskVariantMeasures(),
                'function': self._receive_ask_variant_measures
            },
            {
                'condition': self.ask_variant_gdts,
                'ask': W24AskVariantGDTs(),
                'function': self._receive_ask_variant_gdts
            },
        ]

        # make a short closure to call the correct
        # collector functions
        def make_signal_emitter(item: Dict[str, Any]) -> Callable:
            def func(message: W24TechreadMessage) -> None:
                self.signals.result.emit((item['function'], message))
            return func

        # start with all the hooks that we have from
        # the config
        hooks = [
            Hook(ask=item['ask'], function=make_signal_emitter(item))
            for item in config
            if item['condition']]

        # add a general hook to deal with internal errors
        hooks += [
            Hook(
                message_type=W24TechreadMessageType.ERROR,
                message_subtype=W24TechreadMessageSubtypeError.INTERNAL,
                function=self._receive_error)]

        # add a gneral hook to deal with PROGRESS messages
        hooks += [
            Hook(
                message_type=W24TechreadMessageType.PROGRESS,
                message_subtype=W24TechreadMessageSubtypeProgress.STARTED,
                function=self._receive_progress_started)]

        return hooks

    def _receive_ask_page_thumbnail(
            self,
            message: W24TechreadMessage
    ) -> None:
        """Handle incoming Page Thumbnails

        Args:
            message (W24TechreadMessage): Message with the
                Page Thumbnail
        """
        self.api_feed.add_headline("Page Thumbnail")
        self.api_feed.add_image(message.payload_bytes)
        message.payload_bytes = b"--- binary --"
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()

    def _receive_ask_sheet_thumbnail(
            self,
            message: W24TechreadMessage
    ) -> None:
        """Handle incoming Sheet Thumbnails

        Args:
            message (W24TechreadMessage): Message with the
                Sheet Thumbnail
        """
        self.api_feed.add_headline("Sheet Thumbnail")
        self.api_feed.add_image(message.payload_bytes)
        message.payload_bytes = b"--- binary --"
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()

    def _receive_ask_canvas_thumbnail(
            self,
            message: W24TechreadMessage
    ) -> None:
        """Handle incoming Canvas Thumbnails

        Args:
            message (W24TechreadMessage): Message with the
                Canvas Thumbnail
        """
        self.api_feed.add_headline("Canvas Thumbnail")
        self.api_feed.add_image(message.payload_bytes)
        message.payload_bytes = b"--- binary --"
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()

    def _receive_ask_sectional_thumbnail(
            self,
            message: W24TechreadMessage
    ) -> None:
        """Handle incoming Sectional Thumbnails

        Args:
            message (W24TechreadMessage): Message with the
                Sectional Thumbnail
        """
        self.api_feed.add_headline("Sectional Thumbnail")
        self.api_feed.add_image(message.payload_bytes)

        # store the sectional locally so that we can
        # reuse the image when drawing the measures
        # and gd&ts
        if message.payload_dict is None:
            return

        # write it into the local store
        sectional_id = message.payload_dict.get('sectional_id')
        if sectional_id is not None:
            self.sectional_thumbnails[sectional_id] = message.payload_bytes

        # after we stored it..., censor and print
        message.payload_bytes = b"--- binary --"
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()

    def _receive_ask_variant_measures(
            self,
            message: W24TechreadMessage
    ) -> None:
        """Handle incoming Variant Measures

        Args:
            message (W24TechreadMessage): Message with the
                Variant Measures
        """
        self.api_feed.add_headline("Variant Measures")
        self.api_feed.add_json(message.json())

        # check whether we have an stored version of the sectional
        if message.payload_dict is None:
            return

        sectional_id = message.payload_dict.get('sectional_id')
        if sectional_id is None:
            return

        # now get the measure list from the payload and stop
        # the execution if the payload is None
        measure_list_raw = message.payload_dict.get('measures')
        if measure_list_raw is None or not measure_list_raw:
            return

        # if we have a payload, translate it into usable objects
        measure_list = [W24Measure.parse_obj(m) for m in measure_list_raw]

        # if we really have sectional_bytes and measures, illustrate
        # the results
        sectional_bytes = self.sectional_thumbnails.get(sectional_id)
        if sectional_bytes is not None:
            sectional_bytes_w_measures = event_illustrator\
                .illustrate_sectional_measures(sectional_bytes, measure_list)
            self.api_feed.add_image(sectional_bytes_w_measures)

        # add the table for the measures
        measure_table = W24GuiMeasureTable.from_measure_list(measure_list)
        self.api_feed.add_table(measure_table)
        self.api_feed.add_line()

    def _receive_ask_variant_gdts(
            self,
            message: W24TechreadMessage
    ) -> None:
        """ Handle incoming Variant GD&Ts

        Args:
            message (W24TechreadMessage): Message with the
                Variant GD&Ts
        """
        self.api_feed.add_headline("Variant GD&Ts")
        self.api_feed.add_json(message.json())

        # check whether we have an stored version of the sectional
        if message.payload_dict is None:
            return

        sectional_id = message.payload_dict.get('sectional_id')
        if sectional_id is None:
            return

        # now get the measure list from the payload and stop
        # the execution if the payload is None
        gdt_list_raw = message.payload_dict.get('gdts')
        if gdt_list_raw is None or not gdt_list_raw:
            return

        # if we have a payload, translate it into usable objects
        gdt_list = [W24GDT.parse_obj(m) for m in gdt_list_raw]
        sectional_bytes = self.sectional_thumbnails.get(sectional_id)
        if sectional_bytes is not None:
            sectional_bytes_w_measures = event_illustrator.\
                illustrate_sectional_gdts(
                    sectional_bytes, gdt_list)
            self.api_feed.add_image(sectional_bytes_w_measures)

        # add the table for the measures
        measure_table = W24GuiGdtTable.from_gdt_list(gdt_list)
        self.api_feed.add_table(measure_table)
        self.api_feed.add_line()

    def _receive_error(
            self,
            message: W24TechreadMessage
    ) -> None:
        """ Handle Incomeing errors

        Args:
            message (W24TechreadMessage): Message with the
                Error message
        """
        self.api_feed.add_headline("ERROR")
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()

    def _receive_progress_started(
            self,
            message: W24TechreadMessage
    ) -> None:
        """ Handle Incomeing errors

        Args:
            message (W24TechreadMessage): Message with the
                Progress Message
        """
        self.api_feed.add_headline("Progress")
        self.api_feed.add_json(message.json())
        self.api_feed.add_line()


def main() -> None:
    app = QApplication(sys.argv)

    # Dear Garbadge collector,
    # We are truely grateful for your work.
    # Please do not collect this object though.
    # We still need it.
    # Sincerely yours,
    # Dr. Jay
    _ = W24Gui()

    # run
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()




from typing import List, Dict, Any, Callable

from PyQt5.QtCore import (QObject, QRunnable, pyqtSlot, pyqtSignal)


class W24GuiWorker(QRunnable):
    """ Small worker class to execute non-blocking calls in
    the background. Used to call the API while keeping the
    GUI responsive
    """

    def __init__(
            self,
            function: Callable,
            *args: List[Any],
            **kwargs: Dict[Any, Any]
    ) -> None:
        """ Initiate a new 'thread' that calls
        function when run() is executed

        Args:
            function: funcation to be called
        """
        super(W24GuiWorker, self).__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs

    @pyqtSlot()  # type: ignore # There is little we can do about that
    def run(self) -> None:
        """
        Run the predefined fucntion
        """
        if self.function is not None:
            self.function(*self.args, **self.kwargs)


class W24GuiWorkerSignals(QObject):
    """ Defines the signals available from a running worker thread.
    """
    result = pyqtSignal(object)
    completion = pyqtSignal()



""" Defintions of all objects required to communicate with
the W24 Techread API.
"""
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import UUID4, BaseModel, Field, HttpUrl, Json, validator

from werk24._version import __version__

from .ask import W24Ask, W24AskType, deserialize_ask


class W24TechreadAction(str, Enum):
    """ List of supported actions by the Techread API
    """
    INITIALIZE = "INITIALIZE"
    READ = "READ"


class W24TechreadCommand(BaseModel):
    """ Command that is sent from the client to the Server
    """
    action: W24TechreadAction
    message: Json


class W24TechreadMessageType(str, Enum):
    """ Message Type of the message that is sent
    from the server to the client in response to
    a request.
    """
    ASK = "ASK"
    ERROR = "ERROR"  # !!! DEPRECATED
    PROGRESS = "PROGRESS"
    REJECTION = "REJECTION"


class W24TechreadMessageSubtypeError(str, Enum):
    """ Message Subtype for the MessageType: ERROR

    !!! danger
        The SubtypeError is deprecated in favor of
        the exceptions attribute that is attached to
        the W24AskMessage.
    """
    UNSUPPORTED_DRAWING_FILE_FORMAT = "UNSUPPORTED_DRAWING_FILE_FORMAT"
    INTERNAL = "INTERNAL"
    TIMEOUT = "TIMEOUT"


class W24TechreadMessageSubtypeRejection(str, Enum):
    """ Message Subtype for the MessageType: REJECTION
    """
    COMPLEXITY_EXCEEDED = "COMPLEXITY_EXCEEDED"
    PAPER_SIZE_LIMIT_EXCEEDED = "PAPER_SIZE_LIMIT_EXCEEDED"


class W24TechreadMessageSubtypeProgress(str, Enum):
    """ Message Subtype for the MessageType: PROGRESS
    """
    INITIALIZATION_SUCCESS = "INITIALIZATION_SUCCESS"
    COMPLETED = "COMPLETED"
    STARTED = "STARTED"


W24TechreadMessageSubtypeAsk = W24AskType
""" The MessageType: ASK will return the subtypes
defined in W24AskTypes
"""

W24TechreadMessageSubtype = Union[
    W24TechreadMessageSubtypeError,
    W24TechreadMessageSubtypeProgress,
    W24TechreadMessageSubtypeAsk]
""" Shorthand to summorize all the supported
MessageTypes
"""


class W24TechreadExceptionType(str, Enum):
    """ List of all the error types that can possibly
    be associated to the error type.
    """

    DRAWING_FILE_FORMAT_UNSUPPORTED = "DRAWING_FILE_FORMAT_UNSUPPORTED"
    """ The Drawing was submitted in a file format that is not supproted
    by the API at this stage.
    """

    DRAWING_FILE_SIZE_TOO_LARGE = "DRAWING_FILE_SIZE_TOO_LARGE"
    """ The Drawing file size exceeded the limit
    """

    DRAWING_RESOLUTION_TOO_LOW = "DRAWING_RESOLUTION_TOO_LOW"
    """ The resolution (dots per inch) was too low to be
    processed
    """

    DRAWING_NOISE_TOO_HIGH = "DRAWING_NOISE_TOO_HIGH"
    """ The amount of noise on the drawing was too hight for us
    to understand the drawing
    """

    DRAWING_CONTENT_NOT_UNDERSTOOD = "DRAWING_CONTENT_NOT_UNDERSTOOD"
    """ The file you submitted as drawing might not actually
    be a drawing
    """

    DRAWING_PAPER_SIZE_TOO_LARGE = "DRAWING_PAPER_SIZE_TOO_LARGE"
    """ The paper size is larger that the allowed paper size
    """

    DRAWING_MEASURE_SYSTEM_INCOMPLETE = "DRAWING_MEASURE_SYSTEM_INCOMPLETE"
    """ The file you submitted contains too few measures to be
    manufacturable
    """

    MODEL_FILE_FORMAT_UNSUPPORTED = "MODEL_FILE_FORMAT_UNSUPPORTED"
    """ The Model was submitted in a file format that is not supported
    by the API at this stage.
    """

    MODEL_FILE_SIZE_TOO_LARGE = "MODEL_FILE_SIZE_TOO_LARGE"
    """ The Model fiel size exceeded the limit
    """

    SUB_ACCOUNT_ACCESS_DENIED = "SUB_ACCOUNT_ACCESS_DENIED"
    """Raised when the sub_account does not belong to the
    main account.
    """

    SUB_ACCOUNT_NO_BALANCE = "SUB_ACCOUNT_NO_BALANCE"
    """Raised when the sub account is pre-paid and the budget
    is exhausted.
    """


class W24TechreadExceptionLevel(str, Enum):
    """ Severity level for the Error

    !!! note
        This is defined for future-compatibility.
        The only value that is currently used is ERROR.
        The INFO level will follow shortly.
    """

    ERROR = "ERROR"
    """ Set when the processing was stopped
    """

    INFO = "INFO"
    """ Set when the process was completed successfully,
    but we want to bring something to your awareness
    (e.g,. that you are using a feature that will soon
    be deprecated)
    """


class W24TechreadException(BaseModel):
    """ Error message that accompanies the W24TechreadMessage
    if an error occured.

    Attributes:
        exception_level: Error level indicating the severity of the error

        exception_type: Error Type that allows the API-user to translate
            the message to a user-info.
    """

    exception_level: W24TechreadExceptionLevel

    exception_type: W24TechreadExceptionType


class W24TechreadBaseResponse(BaseModel):
    """BaseFormat for messages returned by the server.

    Attributes:
        exceptions (List[W24TechreadException]): List of exceptions
            that occured during the processing.

    """
    exceptions: List[W24TechreadException] = []

    @property
    def is_successful(self) -> bool:
        """Check whether an exception of the ERROR level was returned.

        Otherwise return True.

        Returns: True if none of the attached exceptions
            has the Exception Level ERROR. False otherwise.
        """
        return not any(
            e.exception_level == W24TechreadExceptionLevel.ERROR
            for e in self.exceptions
        )


class W24TechreadMessage(W24TechreadBaseResponse):
    """ Message format for messages that are sent
    from the server to the client.

    Attributes:
        request_id: unique UUID4 identifier that is generated by the
            server to identify the request

        message_type: Main Message Type (see W24TechreadMessageType)

        message_subtype: Message SubType (see W24TechreadMessageSubtype)

        page_number: Page number that the response belongs to. Starts at
            zero.

        payload_dict: Payload dictionary containing the response
            as dict. The MessageType/Subtype will tell the
            interpreter how to turn the payload back into
            the corresponding object

        payload_url: For binary data, the API will return a download
            url which carries the data. This allows us to transfer
            larger images etc.

        payload_bytes:  Binary reference of the payload. This will only
            become available after the client has downloaded the
            payload_url.


    """
    request_id: UUID4

    message_type: W24TechreadMessageType

    message_subtype: W24TechreadMessageSubtype

    page_number: int = 0

    payload_dict: Optional[Dict] = None

    payload_url: Optional[HttpUrl] = None

    payload_bytes: Optional[bytes] = None


class W24TechreadRequest(BaseModel):
    """ Definition of a W24DrawingReadRequest containing
    all the asks (i.e., things you want to learn about
    the technical drawing).

    Attributes:

        asks: List of asks

        development_key: The development_key is used for internal purposes.
            It wil give you access to pre-release versions of our software.
            You will only understand the details if you...

        client_version: Current version of the client. For backward
            compatibility, this defaults to 'legacy'

        max_pages: Maximum number of pages that shall be processed.

        drawing_filename (Optional[str]): Optional filename

        sub_account (Optional[UUID4]): Sub-account that this request should
            be attributed to. Sub-accounts allow you to keep the requests
            of multiple of your customers separate.
    """
    asks: List[W24Ask] = []

    development_key: Optional[str] = None

    client_version = __version__

    max_pages: int = 1

    drawing_filename: Optional[str] = None

    sub_account: Optional[UUID4] = None

    @validator('asks', pre=True)
    def ask_list_validator(
        cls,
        raw: List[Dict[str, Any]]
    ) -> List[W24Ask]:
        """ Validator to de-serialize the asks. The de-serialization
        is based on the ask_type attribute of the object. Pydantic
        does not support this out-of-the box

        Args:
            raw (Dict[str, Any]): Raw json of the asks list

        Returns:
            List[W24AskUnion]: List of deserialized Asks
        """
        return [deserialize_ask(a) for a in raw]


class W24PresignedPost(BaseModel):
    """ Details of the presigned post that allow you to upload
    a file to our file system.

    Attributes:
        url: Url to which the request shall be sent
        fields_: Dictionary of fields

    """

    url: HttpUrl

    fields_: Dict[str, str] = Field(alias='fields', default={})


class W24TechreadInitResponse(W24TechreadBaseResponse):
    """ API response to the Initialize request

    Attributes:
        drawing_presigned_post: Presigned Post for uploading the drawing

        model_presigned_post: Presigned Post for uploading the model

        exceptions (List[W24TechreadException]): List of exceptions that
            occured

    """

    drawing_presigned_post: W24PresignedPost

    model_presigned_post: W24PresignedPost



from decimal import Decimal
from enum import Enum
from typing import List, Optional, Union
from pydantic import  BaseModel, validator

from .feature import W24FeatureModel

class W24GeneralTolerancesStandard(str, Enum):
    """ Enum of all supported
    General Tolerance Standards
    """
    DIN_7168 = "DIN 7168"
    ISO_2768 = "ISO 2768"
    ISO_4759_1 = "ISO 4759-1"
    DEFINED_ON_SHEET = "DEFINED_ON_SHEET"


class W24ToleranceProperty(str, Enum):
    """ Enum of all attributes that can
    be described by general tolerances
    """
    ANGULAR = "ANGULAR"
    FLATNESS = "FLATNESS"
    LINEAR = "LINEAR"
    PERPENDICULARITY = "PERPENDICULARITY"
    RADIUS = "RADIUS"
    RUNOUT = "RUNOUT"
    STRAIGHTNESS = "STRAIGHTNESS"
    SYMMETRY = "SYMMETRY"


class W24GeneralTolerancesPrinciple(str, Enum):
    """ Enum of the supported General Tolerance
    Principles.
    """
    INDEPENDENCE = "INDEPENDENCE"
    ENVELOPE = "ENVELOPE"


class W24ToleranceTableItem(BaseModel):

    @validator('nominal_min', pre=True)
    def nominal_min_validator(  # NOQA
        cls,
        raw: Union[str, float, None]
    ) -> Optional[Decimal]:
        """ Ensure the proper conversion of the the
        Decimal value
        """
        return cls._convert_decimal(raw)

    @validator('nominal_max', pre=True)
    def nominal_max_validator(  # NOQA
        cls,
        raw: Union[str, float, None]
    ) -> Optional[Decimal]:
        """ Ensure the proper conversion of the the
        Decimal value
        """
        return cls._convert_decimal(raw)

    @staticmethod
    def _convert_decimal(
        raw: Union[str, float, None]
    ) -> Optional[Decimal]:
        """ Handle the decimal converstion

        Args:
            raw (Union[str, float, None]): Raw value

        Raises:
            ValueError: raised when the data type
                does not match the understood types

        Returns:
            Optional[Decimal]: None if the raw value
                corresponds to -Infinity, +Infinity or
                NaN. Decimal('0') if Decimal('-0') or
                the true value
        """

        # ensure we are working with
        # the correct type
        if isinstance(raw, str) \
            or isinstance(raw, float)\
                or isinstance(raw, int):
            decimal = Decimal(raw)

        # accept decimal
        elif isinstance(raw, Decimal):
            decimal = raw

        # allow None values
        elif raw is None:
            return None

        # reject the rest
        else:
            data_type = type(raw)
            raise ValueError(f"Unsupported datatype '{data_type}'")

        # interpret the values
        # handle the special values
        if decimal in {
                Decimal("Infinity"),
                Decimal("-Infinity"),
                Decimal("NaN")}:
            return None

        # enforce zero to be positive
        if decimal == Decimal('-0'):
            return Decimal('0')

        # accept the value
        return decimal

    nominal_min: Optional[Decimal]
    nominal_max: Optional[Decimal]
    deviation_min: Optional[Decimal]
    deviation_max: Optional[Decimal]


class W24ToleranceClass(BaseModel):
    """ Tolerance Class which matches an individual attribute
    of the General Tolerances to a tolerance property and
    tolerance table

    Attributes:
        blurb: Tolerance class label for human consumption

        property: Property that is being tolerated

        table: Rows of the tolerance table that correspond
            to the selected tolerance class
    """
    blurb: str

    property: W24ToleranceProperty

    table: List[W24ToleranceTableItem]


class W24GeneralTolerances(W24FeatureModel):
    """ Object representing the General Tolerances indicated
    on the Title Block of the Technical Drawing.
    """

    blurb: str
    """ Blurb of the general tolerances for human consumption
    """

    tolerance_standard: W24GeneralTolerancesStandard
    """ GeneralTolerance Standard that was defined
    in the Drawing
    """

    principle: Optional[W24GeneralTolerancesPrinciple]
    """ Principle that is annotated on the general
    tolerance by "-E" (or the lack of if).

    NOTE: some general tolerance standards do not define
    a default principle. If it is not stated explicitly
    on the drawing, the `principle` value is set to `None`.
    """

    angular_class: Optional[W24ToleranceClass]
    """ Angular toleration class """

    flatness_class: Optional[W24ToleranceClass]
    """ Flatness toleration class """

    straightness_class: Optional[W24ToleranceClass]
    """ Straightness toleration class """

    linear_class: Optional[W24ToleranceClass]
    """ Linear toleration class """

    radius_class: Optional[W24ToleranceClass]
    """ Radius and chamfer toleration class """

    runout_class: Optional[W24ToleranceClass]
    """ Runout toleration class """

    symmetry_class: Optional[W24ToleranceClass]
    """ Symmetry toleration class """

    perpendicularity_class: Optional[W24ToleranceClass]
    """ Perpendicularity toleration class - not defined in DIN7168 """



from typing import Tuple
import abc
from decimal import Decimal
from enum import Enum
from typing import Optional

from pydantic import UUID4, BaseModel

from .feature import W24FeatureModel
from .unit import W24UnitAngle


class W24AngleTolerationType(str, Enum):
    """ Enum listing all supported AngleToleration Type

    !!! note
        Currently we are only supporting the General
        Tolerances indicated on the drawing's Title Block.
        If you need access to more types, e.g.,
        MINIMUM, MAXIMUM, OFF_SIZE, APPROXIMATION, please
        reach out to us.
    """
    GENERAL_TOLERANCES = "GENERAL_TOLERANCES"


class W24AngleToleration(BaseModel, abc.ABC):
    """ Base Class that describes Angle Tolerances

    Attributes:

        toleration_type: Toleration Type  of the
            `W24AngleSize`

        blurb: String representation for human consumption

    !!! caution
        This model will soon be extended to contain the
        toleration information derived from the TitleBlock.
        Be aware that you will need to request the TitleBlock
        to obtain this information.
    """

    toleration_type: W24AngleTolerationType

    blurb: str


class W24AngleSize(BaseModel):
    """ Size of an Angle including its toleration

    Attributes:

        blurb: Blurb for human consumption

        angle: Nominal angle size in [units]

        unit: Angle Unit. Currently only degrees are
            supported.
    """

    blurb: str

    angle: Decimal

    unit: W24UnitAngle = W24UnitAngle.DEGREE


class W24AngleLabel(BaseModel):
    """ Label associated with an Angle indicated
    on the Technical Drawing

    Attributes:

        blurb: Blurb for human consumption

        size: Nominal angle size

        size_toleration: Tolerated deviations
    """

    blurb: str

    size: W24AngleSize

    size_toleration: W24AngleToleration


class W24Angle(W24FeatureModel):
    """ Tolerated Angle detected on a sectional of the
    Technical Drawing

    Attributes:

        angle_id: Unique identifier of the Angle. This
            allows the reference of the angle from another
            object (e.g., a leader)

        vertex: X and y coordinates of the angle's vertex
            normalized by the sectional's width and height.

        ray1: Unit-vector attached to the vertex that indicates
            the angle's start ray

        ray2: Unit-vector attached to the vertex that indicates
            the angle's end ray

        label: AngleLabel associated with the Angle

        confidence: Werk24 calculates an internal confidence
            score for each angle. The score ranges from 0.0
            to 1.0 and can be used to accept the reading
            as-is or queuing it for a manual verification.
    """

    # !!! DEPRECATED
    vertex: Tuple[float, float]

    # !!! DEPRECATED
    ray1: Tuple[float, float]

    # !!! DEPRECATED
    ray2: Tuple[float, float]

    angle_id: Optional[UUID4]

    label: W24AngleLabel

    confidence: float




""" Defintion of all the W24Radius class its support structures
"""

from typing import Optional

from pydantic import UUID4, BaseModel

from .feature import W24FeatureModel
from .size import W24Size, W24SizeTolerance, W24SizeToleranceGeneral
from .unit import W24UnitLength


class W24RadiusLabel(BaseModel):
    """ Radius Label

    Attributes:
        blurb: String representation of the Radius for human consumption

        quantity: Quantity of the annotated radius, e.g., 2 x R4 returns
            quantity=2

        size: Size of the Radius as referred in the drawing.

        size_tolerance: Tolerance details of the Radius. Please keep in
            mind that Radii can carry special tolerances. If none are
            mentioned on the drawing, the general tolerances apply.

        unit: Length units of the size and size_tolerance. In most cases this
            will be be millimeter (METRIC) or inch (IMPERIAL) and be consistent
            for the complete drawing. Exceptions are very rare, but exist.
    """

    blurb: str

    quality: int = 1

    size: W24Size

    size_tolerance: W24SizeTolerance = W24SizeToleranceGeneral()

    unit: Optional[W24UnitLength] = None


class W24Radius(W24FeatureModel):
    """ Radius Feature

    Attributes:
        radius_id: Unique UUID4 identifier. This can be used to provide
            automated feedback about customer changes.

        label: Label of the radius.

        confidence: Werk24 calcualtes an internal confidence score for
            reach radius. Depending on your use-case, you might want
            to consider or discard low-confidence radii. This value
            allows you to do so. The value ranges from 0.0 to 1.0
    """
    radius_id: UUID4

    label: W24RadiusLabel

    confidence: float



from decimal import Decimal
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class W24DepthThroughType(str, Enum):
    """Through Type specified on ASME drawings.
    """
    THRU = "THRU"
    THRU_ALL = "THRU_ALL"
class W24Depth(BaseModel):
    """ Depth object that describes the details of the
    depth of a drilling or thread

    Attributes:
        blurb: String for human consumption

        depth: Depth of the drilling or thread in the units
            of the parent object
    """
    blurb: str

    depth: Decimal

    through_type: Optional[W24DepthThroughType] = None



import os
from typing import Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QImage, QPixmap
from PyQt5.QtWidgets import (QFrame, QGridLayout, QLabel, QTableWidget,
                             QVBoxLayout)
# from werk24.gui.json_editor import W24GuiJsonEditor
from werk24.gui.style import ft_headline


class QHLine(QFrame):
    def __init__(self) -> None:
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class W24GuiEventFeed(QVBoxLayout):
    """ API Event feed listing all the
    responses that we obtained from the
    API
    """

    def add_welcome(self) -> None:
        """ Add a welcome message that is shown to the
        user before they submitted the first request
        """
        path_cur = os.path.dirname(os.path.abspath(__file__))
        path_logo = os.path.join(
            path_cur,
            "..",
            "assets",
            "images",
            "logo-625.png")

        qimage = QImage(path_logo)
        pixmap = QPixmap(qimage)
        pixmap_label = QLabel()
        pixmap_label.setPixmap(pixmap)

        grid = QVBoxLayout()
        grid.addWidget(pixmap_label)
        grid.addWidget(
            QLabel("Thank you for using W24 Services.\n"
                   "STEP 1: Use the checkboxes on the left to decide "
                   "which asks you want to submit to the API.\n"
                   "STEP 2: Pick the file you want to analyse"))
        self.addStretch(1)

        frame = QFrame()
        frame.setLayout(grid)
        self.addWidget(frame)

    def clear(self) -> None:
        """ Remove all widgets in the event feed
        """
        # for i in reversed(range(self.count())):
        #     try:
        #         print(type(self.itemAt(i).widget()))
        #         #self.itemAt(i).widget().setParent(None)
        #         self.removeWidget(self.itemAt(i).widget())
        #     except AttributeError:
        #         pass
        # print("AS")
        pass

    def add_json(
            self,
            json: str
    ) -> None:
        """ Add a JSON string to the feed

        Args:
            json (str): Json as string
        """
        text = QLabel(json)
        # json_editor = W24GuiJsonEditor()
        # json_editor.setText(json)
        self.addWidget(text)

    def add_headline(
            self,
            headline: str
    ) -> None:
        """ ADd a headline to the feed

        Args:
            headline (str): Headline as string
        """
        lb_headline = QLabel(headline)
        lb_headline.setFont(ft_headline)
        self.addWidget(lb_headline)

    def add_table(
            self,
            table: QTableWidget
    ) -> None:
        """ Add a table to the feed.
        This would probably be a result table for the measures,
        the gd&ts or any future features

        Args:
            table (QTableWidget): Table in question
        """
        self.addWidget(table)

    def add_image(
            self,
            payload_bytes: Optional[bytes]
    ) -> None:
        """ Add an image to the feed

        Args:
            payload_bytes (Optional[bytes]): image in bytes
        """

        # if the payload is None, thre is no point
        if not payload_bytes:
            return

        # otherwise load the image
        qimage = QImage()
        qimage.loadFromData(payload_bytes)

        # rescale it
        pixmap = QPixmap(qimage)
        # pixmap = pixmap.scaled(1024, 1024, Qt.KeepAspectRatio)

        # turn it into a label
        pixmap_label = QLabel()
        pixmap_label.setPixmap(pixmap)

        # and add
        self.addWidget(pixmap_label)

    def add_line(
            self
    ) -> None:
        """ Add a simple separation line
        """
        self.addWidget(QHLine())



import io
import os
from typing import List, Tuple

from PIL import Image, ImageDraw, ImageFont
from werk24.models.gdt import W24GDT, W24GDTCharacteristic
from werk24.models.measure import W24Measure

path_cur = os.path.dirname(os.path.abspath(__file__))
path_font = os.path.join(
    path_cur,
    "..",
    "assets",
    "fonts",
    "STIX2Text-Regular.otf")
font = ImageFont.truetype(path_font, 30)


def _load_image(image_bytes: bytes) -> Tuple[Image.Image, ImageDraw.ImageDraw]:
    img = Image.open(io.BytesIO(image_bytes)).convert('RGB')
    draw = ImageDraw.Draw(img)
    return img, draw


def illustrate_sectional_gdts(
        sectional_bytes: bytes,
        gdts: List[W24GDT]
) -> bytes:

    # open the image from the bytes buffer
    # and dervice the height and width
    img, draw = _load_image(sectional_bytes)
    width, height = img.size

    for cur_gdt in gdts:

        # make a shorter handle for the polygon, and
        # make sure that the polygon does exist and
        # has at least 4 points
        poly = cur_gdt.bounding_polygon
        if not poly or len(poly) < 4:
            continue

        # then draw the polygon onto the image
        for pt1, pt2 in zip(poly, poly[1:] + [poly[0]]):
            draw.line(
                [pt1[0] * width, pt1[1] * height,
                 pt2[0] * width, pt2[1] * height],
                fill=(0, 255, 0),
                width=4)

        # now... we want to have a measure label
        # that is on the center coordinate of the
        # measure
        left = min([float("inf")] + [pt[0] for pt in poly])
        top = min([float("inf")] + [pt[1] for pt in poly])
        right = max([0.0] + [pt[0] for pt in poly])
        bottom = max([0.0] + [pt[1] for pt in poly])
        x_center = (left + right) / 2 * width
        y_center = (top + bottom) / 2 * height
        if left == float("inf") or top == float("inf"):
            continue

        # replace the special characters by something
        # that is more easily digestable
        blurb = cur_gdt.frame.blurb
        trans_table = str.maketrans(
            {k.value: k.name
             for k in W24GDTCharacteristic
             if len(k.value) == 1})
        blurb = blurb.translate(trans_table)

        # cur_measure.label.blurb
        draw.text(
            [x_center, y_center],
            blurb,
            fill=(0, 200, 0),
            font=font)

    return img.tobytes("jpeg", "RGB")


def illustrate_sectional_measures(
        sectional_bytes: bytes,
        measures: List[W24Measure]
) -> bytes:

    # open the image from the bytes buffer
    # and dervice the height and width
    img, draw = _load_image(sectional_bytes)
    width, height = img.size

    for cur_measure in measures:

        # draw a line that corresponds to the
        # measure line
        cur_line = cur_measure.line
        draw.line(
            [cur_line[0][0] * width,
             cur_line[0][1] * height,
             cur_line[1][0] * width,
             cur_line[1][1] * height],
            fill=(0, 255, 0),
            width=4)

        # now... we want to have a measure label
        # that is on the center coordinate of the
        # measure
        x_center = (cur_line[0][0] + cur_line[1][0]) / 2 * width
        y_center = (cur_line[0][1] + cur_line[1][1]) / 2 * height
        # cur_measure.label.blurb
        draw.text(
            [x_center, y_center],
            cur_measure.label.blurb,
            fill=(0, 200, 0),
            font=font)

    return img.tobytes("jpeg", "RGB")



import asyncio
from typing import List

from werk24 import Hook, W24TechreadClient


async def w24_read_async(
    document_bytes: bytes,
    hooks:  List[Hook],
    max_pages: int = 1
) -> None:
    """ Convenient handler to submit an asyncronous request to the
    Werk24 TechRead API.

    Args:
        document_bytes (bytes): Document bytes
        hooks (List[Hook]): List of Hooks that you want to
            submit.
    """
    async with W24TechreadClient.make_from_env() as session:
        await session.read_drawing_with_hooks(document_bytes, hooks, max_pages)


def w24_read_sync(
    document_bytes: bytes,
    hooks:  List[Hook],
    max_pages: int = 1
) -> None:
    """ Convenient handler to submit a syncronous request to the
    Werk24 TechRead API.

    Args:
        document_bytes (bytes): Document bytes
        hooks (List[Hook]): List of Hooks that you want to
            submit.
    """
    asyncio.run(w24_read_async(document_bytes, hooks, max_pages))



"""Color Models.
"""
from pydantic import BaseModel
from enum import Enum


class W24ColorSystem(str, Enum):
    """Color Systems.

    Multiple color systems exist. Ranging from pure
    textual descriptions to highly-standardized systems.

    In the current implementation we do not make these
    colors comparable - their appearance might even
    vary across materials.

    Rather, we start with the simple implementation of
    the RAL system. At later stages, other color system
    could follow.
    """
    BLURB = "BLURB"
    RAL = "RAL"
    # FED_STD_595 = "FED_STD_595"
    # CIE = "CIE"
    # PANTONE = "PANTONE"
    # HKS = "HKS"
    # NCS = "NCS"


class W24Color(BaseModel):
    """W24 Model for Colors.

    Attributes:
        blurb (str): String representation for human consumption

        system: Color system used to specify
            the color.

        designation: Color designation in the corresponding
            system.
    """
    blurb: str
    system: W24ColorSystem
    designation: str



""" Material Models
"""
from enum import Enum
from typing import Tuple, Optional, Any
from werk24.models.feature import W24FeatureModel


class W24MaterialCategory1(str, Enum):
    FERROUS_ALLOY = "FERROUS_ALLOY"
    NONFERROUS_ALLOY = "NONFERROUS_ALLOY"
    POLYMER = "POLYMER"
    CERAMIC = "CERAMIC"
    COMPOSITE = "COMPOSITE"
    ORGANIC = "ORGANIC"


class W24MaterialCategory2(str, Enum):

    # FERROUS ALLOYS
    STEEL = "STEEL"
    IRON = "IRON"
    MAGNETIC_OR_ELECTRICAL_MATERIAL = "MAGNETIC_OR_ELECTRICAL_MATERIAL"
    SINTERED_POWDEDERD_METAL = "SINTERED_POWDEDERD_METAL"
    WELDING_FILLER_MATERIAL = "WELDING_FILLER_MATERIAL"

    # NONFERROUS ALLOY
    ALUMINUM = "ALUMINUM"
    ANTIMONY = "ANTIMONY"
    BERYLLIUM = "BERYLLIUM"
    BISMUTH = "BISMUTH"
    CADMIUM = "CADMIUM"
    CHROMIUM = "CHROMIUM"
    COBALT = "COBALT"
    COPPER = "COPPER"
    GALLIUM = "GALLIUM"
    LEAD = "LEAD"
    MAGNESIUM = "MAGNESIUM"
    MANGANESE = "MANGANESE"
    MERCURY = "MERCURY"
    MOLYBDENUM = "MOLYBDENUM"
    NICKEL = "NICKEL"
    NIOBIUM = "NIOBIUM"
    PLATINUM = "PLATINUM"
    REAR_EARTHS = "REAR_EARTHS"
    SELENIUM = "SELENIUM"
    SILICON = "SILICON"
    THALLIUM = "THALLIUM"
    TIN = "TIN"
    TITANIUM = "TITANIUM"
    TUNGSTEN = "TUNGSTEN"
    VANADIUM = "VANADIUM"
    ZINC = "ZINC"
    ZIRCONIUM = "ZIRCONIUM"

    # POLYMER
    THERMOPLAST = "THERMOPLAST"
    THERMOSET = "THERMOSET"
    THERMOPLASTIC_ELASTOMER = "THERMOPLASTIC_ELASTOMER"
    THERMOSET_ELASTOMER = "THERMOSET_ELASTOMER"
    POLYMER_FOAM = "POLYMER_FOAM"

    # CERAMIC
    REFRACTORY = "REFRACTORY"
    TECHNICAL_CERAMIC = "TECHNICAL_CERAMIC"
    CERAMIC_FOAM = "CERAMIC_FOAM"

    # COMPOSITE
    MATRIX = "MATRIX"
    REINFORCEMENT = "REINFORCEMENT"

    # ORGANIC
    WOOD = "WOOD"


class W24MaterialCategory3(str, Enum):

    # FERROUS_ALLOY / STEEL
    STRUCTURAL_OR_CONSTRUCTIONAL_STEEL = "STRUCTURAL_OR_CONSTRUCTIONAL_STEEL"
    STAINLESS_STEEL = "STAINLESS_STEEL"
    TOOL_STEEL = "TOOL_STEEL"
    CAST_STEEL = "CAST_STEEL"

    # FERROUS_ALLOY / IRON
    CAST_IRON = "CAST_IRON"
    FERROALLOY = "FERROALLOY"

    # POLYMER / THERMOPLAST
    ABS = "ABS"
    ABS_PC = "ABS_PC"
    ABS_PLUS_PA = "ABS_PLUS_PA"
    ABS_PLUS_PA6 = "ABS_PLUS_PA6"
    ABS_PLUS_PA66 = "ABS_PLUS_PA66"
    ABS_PLUS_PBT = "ABS_PLUS_PBT"
    ABS_PLUS_PC = "ABS_PLUS_PC"
    ABS_PLUS_PET = "ABS_PLUS_PET"
    ABS_PLUS_PMMA = "ABS_PLUS_PMMA"
    ABS_PLUS_PVC = "ABS_PLUS_PVC"
    ABS_PLUS_PVC_C = "ABS_PLUS_PVC_C"
    ABS_PLUS_SAN = "ABS_PLUS_SAN"
    ABS_PLUS_TPE = "ABS_PLUS_TPE"
    ABS_PLUS_TPES = "ABS_PLUS_TPES"
    ABS_PLUS_TPU = "ABS_PLUS_TPU"
    ABS_PP = "ABS_PP"
    ACPES = "ACPES"
    AES = "AES"
    AES_PLUS_PC = "AES_PLUS_PC"
    ASA = "ASA"
    ASA_PC = "ASA_PC"
    ASA_PLUS_ABS = "ASA_PLUS_ABS"
    ASA_PLUS_AES = "ASA_PLUS_AES"
    ASA_PLUS_MSAN = "ASA_PLUS_MSAN"
    ASA_PLUS_PA = "ASA_PLUS_PA"
    ASA_PLUS_PC = "ASA_PLUS_PC"
    ASA_PLUS_PMMA = "ASA_PLUS_PMMA"
    ASA_PLUS_SAN = "ASA_PLUS_SAN"
    ASA_PLUS_TPE = "ASA_PLUS_TPE"
    BMI = "BMI"
    BVOH = "BVOH"
    CAB = "CAB"
    CAP = "CAP"
    CE = "CE"
    COC = "COC"
    CP = "CP"
    CPE = "CPE"
    CPT = "CPT"
    E_P = "E_P"
    EA = "EA"
    EAA = "EAA"
    EBA = "EBA"
    EBACO = "EBACO"
    EC = "EC"
    ECTFE = "ECTFE"
    EEA = "EEA"
    EFEP = "EFEP"
    EMAA = "EMAA"
    EMAAA = "EMAAA"
    ENBAGMA = "ENBAGMA"
    ETFE = "ETFE"
    EVA = "EVA"
    EVACO = "EVACO"
    EVOH = "EVOH"
    FEP = "FEP"
    HDPE = "HDPE"
    HIPP = "HIPP"
    HIPS = "HIPS"
    IONOMER_RESIN = "IONOMER_RESIN"
    LCP = "LCP"
    LCP_PLUS_PPS = "LCP_PLUS_PPS"
    LDPE = "LDPE"
    LDPE_EVA = "LDPE_EVA"
    LDPE_LLDPE = "LDPE_LLDPE"
    LLDPE = "LLDPE"
    LMDPE = "LMDPE"
    MABS = "MABS"
    MBS = "MBS"
    MDPE = "MDPE"
    MFA = "MFA"
    PA = "PA"
    PA_MACM10_1010 = "PA_MACM10_1010"
    PA_MACM12 = "PA_MACM12"
    PA_MACM12_PLUS_PA_12 = "PA_MACM12_PLUS_PA_12"
    PA_MXD6 = "PA_MXD6"
    PA_MXD6_MXDI = "PA_MXD6_MXDI"
    PA_PACM12 = "PA_PACM12"
    PA_PLUS_HDPE = "PA_PLUS_HDPE"
    PA_PLUS_PA66_6 = "PA_PLUS_PA66_6"
    PA_PLUS_PE = "PA_PLUS_PE"
    PA_PLUS_PP = "PA_PLUS_PP"
    PA_PLUS_PPA = "PA_PLUS_PPA"
    PA_PLUS_SPS = "PA_PLUS_SPS"
    PA_PLUS_TPE = "PA_PLUS_TPE"
    PA_TPE = "PA_TPE"
    PA1010 = "PA1010"
    PA1012 = "PA1012"
    PA10T_X = "PA10T_X"
    PA11 = "PA11"
    PA12 = "PA12"
    PA12_MACMI = "PA12_MACMI"
    PA12_MACMI_PLUS_PA12 = "PA12_MACMI_PLUS_PA12"
    PA410 = "PA410"
    PA46 = "PA46"
    PA46_PLUS_PA6 = "PA46_PLUS_PA6"
    PA4T = "PA4T"
    PA510 = "PA510"
    PA6 = "PA6"
    PA6_12 = "PA6_12"
    PA6_3T = "PA6_3T"
    PA6_66_12 = "PA6_66_12"
    PA6_66_136 = "PA6_66_136"
    PA6_66_610 = "PA6_66_610"
    PA6_69 = "PA6_69"
    PA6_6I = "PA6_6I"
    PA6_6T = "PA6_6T"
    PA6_IPDI = "PA6_IPDI"
    PA6_PA12 = "PA6_PA12"
    PA6_PLUS_ASA = "PA6_PLUS_ASA"
    PA6_PLUS_PA_6I_6T = "PA_6_PLUS_PA_6I_6T"
    PA6_PLUS_PA12 = "PA6_PLUS_PA12"
    PA6_PLUS_PA12_X = "PA6_PLUS_PA12_X"
    PA6_PLUS_PA66 = "PA6_PLUS_PA66"
    PA6_PLUS_PA66_6 = "PA6_PLUS_PA66_6"
    PA6_PLUS_PE = "PA6_PLUS_PE"
    PA6_PLUS_PP = "PA6_PLUS_PP"
    PA610 = "PA610"
    PA612 = "PA612"
    PA66 = "PA66"
    PA66_6 = "PA66_6"
    PA66_610 = "PA66_610"
    PA66_PLUS_PA610 = "PA66_PLUS_PA610"
    PA66_PLUS_PA612 = "PA66_PLUS_PA612"
    PA66_PLUS_PA6I = "PA66_PLUS_PA6I"
    PA66_PLUS_PA6I_6T = "PA66_PLUS_PA6I_6T"
    PA66_PLUS_PA6I_X = "PA66_PLUS_PA6I_X"
    PA66_PLUS_PE = "PA66_PLUS_PE"
    PA66_PLUS_PP = "PA66_PLUS_PP"
    PA66_PLUS_PPA = "PA66_PLUS_PPA"
    PA69 = "PA69"
    PA6I = "PA6I"
    PA6I_6T = "PA6I_6T"
    PA6I_X = "PA6I_X"
    PA6T = "PA6T"
    PA6T_66 = "PA6T_66"
    PA6T_XT_PLUS_PA6T_66 = "PA6T_XT_PLUS_PA6T_66"
    PA6T_6I = "PA6T_6I"
    PA6T_6I_66 = "PA6T_6I_66"
    PA6T_MPMDT = "PA6T_MPMDT"
    PA6T_XT = "PA6T_XT"
    PA9T = "PA9T"
    PAEK = "PAEK"
    PAI = "PAI"
    PAMXD6_PA66 = "PAMXD6_PA66"
    PAO = "PAO"
    PAR = "PAR"
    PARA = "PARA"
    PB = "PB"
    PBAT = "PBAT"
    PBAT_PLUS_PLA = "PBAT_PLUS_PLA"
    PBI = "PBI"
    PBS = "PBS"
    PBT = "PBT"
    PBT_ASA = "PBT_ASA"
    PBT_PC = "PBT_PC"
    PBT_PLUS_ASA = "PBT_PLUS_ASA"
    PBT_PLUS_PE = "PBT_PLUS_PE"
    PBT_PLUS_PET = "PBT_PLUS_PET"
    PBT_PLUS_PET_PLUS_ASA = "PBT_PLUS_PET_PLUS_ASA"
    PBT_PLUS_PETG = "PBT_PLUS_PETG"
    PBT_PLUS_PPE = "PBT_PLUS_PPE"
    PBT_PLUS_PS = "PBT_PLUS_PS"
    PBT_PLUS_SAN = "PBT_PLUS_SAN"
    PC = "PC"
    PC_PET = "PC_PET"
    PC_PLUS_HIPS = "PC_PLUS_HIPS"
    PC_PLUS_MBS = "PC_PLUS_MBS"
    PC_PLUS_PBT = "PC_PLUS_PBT"
    PC_PLUS_PCT = "PC_PLUS_PCT"
    PC_PLUS_PET = "PC_PLUS_PET"
    PC_PLUS_PET_PLUS_PBT = "PC_PLUS_PET_PLUS_PBT"
    PC_PLUS_PLA = "PC_PLUS_PLA"
    PC_PLUS_PMMA = "PC_PLUS_PMMA"
    PC_PLUS_PPC = "PC_PLUS_PPC"
    PC_PLUS_PS = "PC_PLUS_PS"
    PC_PLUS_PTFE = "PC_PLUS_PTFE"
    PC_PLUS_SAN = "PC_PLUS_SAN"
    PC_PLUS_TPES = "PC_PLUS_TPES"
    PC_PLUS_TPU = "PC_PLUS_TPU"
    PC_PPC = "PC_PPC"
    PCL = "PCL"
    PCT = "PCT"
    PCTA = "PCTA"
    PCTFE = "PCTFE"
    PCTG = "PCTG"
    PCTG_PLUS_PC = "PCTG_PLUS_PC"
    PE = "PE"
    PE_HMW = "PE_HMW"
    PE_HMWHD = "PE_HMWHD"
    PE_PLUS_HIPS = "PE_PLUS_HIPS"
    PE_UHMW = "PE_UHMW"
    PEEK = "PEEK"
    PEI = "PEI"
    PEI_PLUS_PCE = "PEI_PLUS_PCE"
    PEK = "PEK"
    PEKEKK = "PEKEKK"
    PEKK = "PEKK"
    PEN = "PEN"
    PESU = "PESU"
    PET = "PET"
    PET_G = "PET_G"
    PET_PLUS_PA6 = "PET_PLUS_PA6"
    PET_PLUS_PA66 = "PET_PLUS_PA66"
    PEX_B = "PEX_B"
    PFA = "PFA"
    PFSA_PTFE = "PFSA_PTFE"
    PGA = "PGA"
    PHA = "PHA"
    PHB = "PHB"
    PHBV = "PHBV"
    PI = "PI"
    PIB = "PIB"
    PK = "PK"
    PLA = "PLA"
    PLA_PCL = "PLA_PCL"
    PLA_PEG = "PLA_PEG"
    PLA_PHA = "PLA_PHA"
    PLA_PLUS_ABS = "PLA_PLUS_ABS"
    PLA_PLUS_HDPE = "PLA_PLUS_HDPE"
    PLA_PLUS_PHB = "PLA_PLUS_PHB"
    PLA_PLUS_PMMA = "PLA_PLUS_PMMA"
    PLGA = "PLGA"
    PMMA = "PMMA"
    PMMI = "PMMI"
    PMP = "PMP"
    PMS = "PMS"
    POM = "POM"
    POM_PLUS_MBS = "POM_PLUS_MBS"
    POM_PLUS_PE = "POM_PLUS_PE"
    POM_PLUS_PTFE = "POM_PLUS_PTFE"
    POM_PLUS_PUR = "POM_PLUS_PUR"
    PP = "PP"
    PP_PE = "PP_PE"
    PP_PLUS_EPDM = "PP_PLUS_EPDM"
    PP_PLUS_EPR = "PP_PLUS_EPR"
    PP_PLUS_EVA = "PP_PLUS_EVA"
    PP_PLUS_PE = "PP_PLUS_PE"
    PP_PLUS_PPE_PLUS_PS = "PP_PLUS_PPE_PLUS_PS"
    PPA = "PPA"
    PPC = "PPC"
    PPE = "PPE"
    PPE_PLUS_HIPS = "PPE_PLUS_HIPS"
    PPE_PLUS_PA = "PPE_PLUS_PA"
    PPE_PLUS_PA6 = "PPE_PLUS_PA6"
    PPE_PLUS_PA66 = "PPE_PLUS_PA66"
    PPE_PLUS_PP = "PPE_PLUS_PP"
    PPE_PLUS_PS = "PPE_PLUS_PS"
    PPE_PLUS_PS_PLUS_PA = "PPE_PLUS_PS_PLUS_PA"
    PPE_PLUS_TPE = "PPE_PLUS_TPE"
    PPE_PLUS_TPS_SEBS = "PPE_PLUS_TPS_SEBS"
    PPOX = "PPOX"
    PPS = "PPS"
    PPS_PLUS_PA = "PPS_PLUS_PA"
    PPS_PLUS_PPE = "PPS_PLUS_PPE"
    PPS_PLUS_PTFE = "PPS_PLUS_PTFE"
    PPSU = "PPSU"
    PPSU_PLUS_PSU = "PPSU_PLUS_PSU"
    PS = "PS"
    PS_I = "PS_I"
    PS_PLUS_PE = "PS_PLUS_PE"
    PS_PLUS_PMMA = "PS_PLUS_PMMA"
    PS_PLUS_SPS = "PS_PLUS_SPS"
    PSU = "PSU"
    PSU_PLUS_ABS = "PSU_PLUS_ABS"
    PSU_PLUS_PC = "PSU_PLUS_PC"
    PTFE = "PTFE"
    PTT = "PTT"
    PVAL = "PVAL"
    PVB = "PVB"
    PVC = "PVC"
    PVC_C = "PVC_C"
    PVC_C_PLUS_PVC = "PVC_C_PLUS_PVC"
    PVC_PLUS_NBR = "PVC_PLUS_NBR"
    PVC_PLUS_PMMA = "PVC_PLUS_PMMA"
    PVC_PLUS_PUR = "PVC_PLUS_PUR"
    PVC_U = "PVC_U"
    PVCA = "PVCA"
    PVDC = "PVDC"
    PVDF = "PVDF"
    PVP = "PVP"
    RPS = "RPS"
    SAN = "SAN"
    SB = "SB"
    SMA = "SMA"
    SMA_PLUS_HIPS = "SMA_PLUS_HIPS"
    SMI = "SMI"
    SMMA = "SMMA"
    SPS = "SPS"
    SRP = "SRP"
    TPC_EE_PLUS_PBT = "TPC_EE_PLUS_PBT"
    TPC_ET_PLUS_EMA = "TPC_ET_PLUS_EMA"
    TPC_ET_PLUS_PBT = "TPC_ET_PLUS_PBT"
    TPES = "TPES"
    TPU_PLUS_TPS_PLUS_TPA = "TPU_PLUS_TPS_PLUS_TPA"
    VDF_CTFE = "VDF_CTFE"
    VDF_HFP = "VDF_HFP"
    VINYL_CHLORIDE_BLEND = "VINYL_CHLORIDE_BLEND"
    VLDPE = "VLDPE"
    XLPE = "XLPE"
    XLPO = "XLPO"

    # POLYMER / THERMOSET
    CA = "CA"
    EP = "EP"
    MA = "MA"
    MF = "MF"
    MP = "MP"
    PDAIP = "PDAIP"
    PDAP = "PDAP"
    PUR = "PUR"
    UF = "UF"
    UP = "UP"
    VE = "VE"

    # POLYMER / THERMOPLASTIC_ELASTOMER
    MPR = "MPR"
    PCU = "PCU"
    PEBA = "PEBA"
    SPU = "SPU"
    TPA = "TPA"
    TPC = "TPC"
    TPC_EE = "TPC_EE"
    TPC_ES = "TPC_ES"
    TPC_ET = "TPC_ET"
    TPE = "TPE"
    TPO = "TPO"
    TPO_EB = "TPO_EB"
    TPO_EO = "TPO_EO"
    TPO_EPDM_PLUS_PP = "TPO_EPDM_PLUS_PP"
    TPS = "TPS"
    TPS_SBS = "TPS_SBS"
    TPS_SEBS = "TPS_SEBS"
    TPS_SIS = "TPS_SIS"
    TPU = "TPU"
    TPU_ALES = "TPU_ALES"
    TPU_ALET = "TPU_ALET"
    TPU_ARES = "TPU_ARES"
    TPU_ARET = "TPU_ARET"
    TPU_ES = "TPU_ES"
    TPU_ET = "TPU_ET"
    TPV = "TPV"
    TPV_EPDM = "TPV_EPDM"
    TPV_EPDM_PLUS_PP = "TPV_EPDM_PLUS_PP"
    TPV_IIR_HIIR = "TPV_IIR_HIIR"
    TPV_NBR = "TPV_NBR"
    TSPCU = "TSPCU"
    TSPU = "TSPU"

    # POLYMER / THERMOSET_ELASTOMER
    ACM = "ACM"
    ACSM = "ACSM"
    AEM = "AEM"
    AU = "AU"
    BIIR = "BIIR"
    BR = "BR"
    CIIR = "CIIR"
    CM = "CM"
    CR = "CR"
    CR_NBR = "CR_NBR"
    CR_SBR = "CR_SBR"
    CSM = "CSM"
    ECO = "ECO"
    EPDM = "EPDM"
    EPDM_SBR = "EPDM_SBR"
    EPM = "EPM"
    EPM_EPDM = "EPM_EPDM"
    EPT = "EPT"
    EU = "EU"
    FFKM = "FFKM"
    FKM = "FKM"
    FMQ = "FMQ"
    FVMQ = "FVMQ"
    HNBR = "HNBR"
    IIR = "IIR"
    IR = "IR"
    IR_BR = "IR_BR"
    NBR = "NBR"
    NBR_PVC = "NBR_PVC"
    NBR_SBR = "NBR_SBR"
    NR = "NR"
    PDMS = "PDMS"
    PUMA = "PUMA"
    PVMQ = "PVMQ"
    RET = "RET"
    SBR = "SBR"
    SBR_ACM = "SBR_ACM"
    SBR_BR = "SBR_BR"
    SBR_IR = "SBR_IR"
    SBR_NR = "SBR_NR"
    SI = "SI"
    VMQ = "VMQ"
    VMQ_PVMQ = "VMQ_PVMQ"
    XNBR = "XNBR"
    XNBR_PVC = "XNBR_PVC"

    # COMPOSITE
    CARBON_FIBER_COMPOSITE = "CARBON_FIBER_COMPOSITE"
    GLASS_FIBER_COMPOSITE = "GLASS_FIBER_COMPOSITE"

    # REINFORCEMENT
    FIBER_REINFORCEMENT = "FIBER_REINFORCEMENT"
    STRUCTURAL_COMPOSITE = "STRUCTURAL_COMPOSITE"

    # FOAM / POLYMER_FOAM
    POLYPHENYLENE_OR_POLYSTYRENE = "POLYPHENYLENE_OR_POLYSTYRENE"
    POLYPHENYLENE_OR_POLYSTYRENE_HIPS = "POLYPHENYLENE_OR_POLYSTYRENE_HIPS"
    POLYCARBONATE = "POLYCARBONATE"


class W24Material(W24FeatureModel):
    """W24 Object for Materials.

    Parsed Material object that can either be
    associated to the TitleBlock or derived from
    all the available information (including the
    text on the canvas.

    Attributes:
        blurb: Material Name for human consumption.
            This will typically include the designation
            and the standard.

        standard: Material Standard indicated
            on the technical drawing. This used to be
            an enum - but we now have over 100 supported
            material standards and the number is increasing
            weekly. So a string seems to be more appropriate.

        designation: Name of the material in accordance
            with the material standard. An alternative name
            would be material_designation.

        material_category: Categorization of the material
            following W24MaterialCategoryX Tree.

    """
    blurb: str
    standard: str
    designation: str
    material_category: Tuple[
        Optional[W24MaterialCategory1],
        Optional[W24MaterialCategory2],
        Optional[W24MaterialCategory3],
    ]

    # ! DEPRECATED in version 1.4.0
    material_family: Optional[Any] = None
    material_class: Optional[Any] = None
    material_type: Optional[Any] = None
    material_group: Optional[Any] = None
    material_standard: Any
    material_code: str



from datetime import date

from pydantic import BaseModel, validator


class W24Date(BaseModel):
    """ Date object

    Attributes:
        blurb: Date in ISO 8601 format

        date: Python date object; this will
            automatically be interpreted when
            the object is loaded
    """

    @validator('date', allow_reuse=True)
    def date_validator(cls, v:date) -> date:
        return v.isoformat()

    blurb: str

    date: date



from typing import List, Optional

from pydantic import BaseModel


class W24FeatureCoordinate(BaseModel):
    """ Coordinate point

    Attributes:
        x: x position normalized by the thumbnail's width

        y: y position normalized by the thumbnail's height
    """
    x: float
    y: float


class W24FeaturePosition(BaseModel):
    """ Position of the Feature on the individual thumbnails normalized
    by the width and height of each thumbnail.

    Each features position is indicated as a list of coordinates. If the
        list only has two elements, you are dealing with a line. If it
        has four or more, you are looking at a polygon

    Attributes:
        page: Position of the Feature on the Page thumbnail

        sheet: Position of the Feature on the Sheet thumbnail

        sectional: Position of the Feature on the Sectional thumbnail

    """
    sheet: List[W24FeatureCoordinate]
    canvas: List[W24FeatureCoordinate]
    sectional: List[W24FeatureCoordinate]


class W24FeatureModel(BaseModel):
    """ Base Model for all the features that we might
    extract from the Drawing

    Attributes:
        position: Position of the features on the individual
            thumbnails
    """

    # NOTE: position is optional for the transition period
    position: Optional[W24FeaturePosition] = None





from typing import List, Callable, Dict
from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QHeaderView
from PyQt5.QtCore import QRect
from werk24.models.measure import W24Measure


class W24GuiMeasureTable(QTableWidget):

    @classmethod
    def from_measure_list(
            cls,
            measure_list: List[W24Measure]
    ) -> 'W24GuiMeasureTable':

        # make a simple config file for the current table
        config: Dict[str, Callable] = {
            "Blurb": cls._format_measure_blurb,
            "Quantity": cls._format_measure_quantity,
            "Size": cls._format_measure_size,
            "Tolerance": cls._format_measure_tolerance,
            "Thread": cls._format_measure_thread,
            "Chamfer": cls._format_measure_chamfer,
            "Start": cls._format_measure_start,
            "Stop": cls._format_measure_stop,
            "Confidence": cls._format_measure_confidence
        }

        # make the associated table
        table_widget = W24GuiMeasureTable()
        table_widget.setColumnCount(len(config))
        table_widget.setRowCount(len(measure_list))
        table_widget.verticalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents)

        # make the table content
        table_widget.setHorizontalHeaderLabels(config.keys())
        for i, cur_measure in enumerate(measure_list):
            for j, callable_ in enumerate(config.values()):
                content = callable_(cur_measure)
                table_widget.setItem(i, j, QTableWidgetItem(content))
            table_widget.resizeColumnToContents(i)

        table_widget.setGeometry(0, 0, 200, 200)
        table_widget.resizeRowsToContents()
        return table_widget

    @staticmethod
    def _format_measure_confidence(measure: W24Measure) -> str:
        return f"{measure.confidence}"

    @staticmethod
    def _format_measure_blurb(measure: W24Measure) -> str:
        return f"{measure.label.blurb}"

    @staticmethod
    def _format_measure_quantity(measure: W24Measure) -> str:
        return f"{measure.label.quantity}"

    @staticmethod
    def _format_measure_size(measure: W24Measure) -> str:
        return f"{measure.label.size.blurb}"

    @staticmethod
    def _format_measure_tolerance(measure: W24Measure) -> str:
        return "" if measure.label.size_tolerance is None \
            else f"{measure.label.size_tolerance.blurb}"

    @staticmethod
    def _format_measure_thread(measure: W24Measure) -> str:
        return "" if measure.label.thread is None \
            else f"{measure.label.thread.blurb}"

    @staticmethod
    def _format_measure_chamfer(measure: W24Measure) -> str:
        return "" if measure.label.chamfer is None \
            else f"{measure.label.chamfer.blurb}"

    @staticmethod
    def _format_measure_start(measure: W24Measure) -> str:
        return f"{measure.line[0][0]:.4f} x {measure.line[0][1]:.4f}"

    @staticmethod
    def _format_measure_stop(measure: W24Measure) -> str:
        return f"{measure.line[1][0]:.4f} x {measure.line[1][1]:.4f}"



""" Command Line Interface for W24
"""
import argparse
import asyncio
from dotenv import load_dotenv
import werk24.cli.techread
import werk24.cli.auth

load_dotenv(".werk24")


def main() -> None:

    parser = argparse.ArgumentParser(prog="w24cli")
    subparsers = parser.add_subparsers(
        dest="command",
        help="Subcommand. Currently supported: techread")
    subparsers.required = True

    parser_techread = subparsers.add_parser(
        "techread",
        help="Submit a Technical Drawing to Werk24 for analysis")

    parser_techread.add_argument(
        "input_file",
        help="path to the file that is to be analyzed")

    parser_techread.add_argument(
        "--sub-account",
        help="Sub-Account that the request should be attributed to.",
        action="store",
        dest='sub_account',
        default=None)

    parser_techread.add_argument(
        "--ask-techread-started",
        help="ask for a callback when the techread process has been picked up by a worker",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-page-thumbnail",
        help="ask for a thumbnail for each page in the document",
        action="store_true")

    parser_techread.add_argument(
        "--ask-sheet-thumbnail",
        help="ask for a thumbnail of each sheet in the document",
        action="store_true")

    parser_techread.add_argument(
        "--ask-sheet-anonymization",
        help="ask for an anonymized of each sheet in the document",
        action="store_true")

    parser_techread.add_argument(
        "--ask-canvas-thumbnail",
        help="ask for a thumbnail of each canvas in the document",
        action="store_true")

    parser_techread.add_argument(
        "--ask-sectional-thumbnail",
        help="ask for a thumbnail of each sectional of each sheet in the document",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-cad",
        help="ask for the CAD file of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-external-dimensions",
        help="ask for the External Dimensions of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-gdts",
        help="ask for the GD&Ts of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-measures",
        help="ask for the measures of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-roughnesses",
        help="ask for the Surface Roughnesses of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-radii",
        help="ask for the Radii of each variant",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-variant-thread-elements",
        help="ask for the Thread Elements",  # noqa
        action="store_true")

    parser_techread.add_argument(
        "--ask-titleblock",
        help="ask for the Title Block",  # noqa
        action="store_true")

    parser_auth = subparsers.add_parser(
        "auth",
        help="Interact with the authentication service")

    parser_auth.add_argument(
        '--ask-jwt-token',
        help="Obtain a valid JWT token",
        action="store_true")

    args = parser.parse_args()
    if args.command == "techread":
        asyncio.run(werk24.cli.techread.main(args))

    elif args.command == "auth":
        asyncio.run(werk24.cli.auth.main(args))


if __name__ == "__main__":
    main()



""" Module for all exceptions
"""


class TechreadException(Exception):
    """ Base Exception for all Exceptions raised by
    the Techread functionality

    """


class UnauthorizedException(TechreadException):
    """ Exception that is raised when
    (i) the response code is 403 - Unauthorized, or
    (ii) the requested action was forbidden by the gateway
    """


class UnknownException(TechreadException):
    """ A developer's favorite exception
    """


class RequestTooLargeException(TechreadException):
    """ Raised RequestTooLargeException the request exceeds the maximal
    request size (at the time of writing 6MB).
    """


class ServerException(TechreadException):
    """ Exception that is raised, when the server responded in an unexpected
    way.
    """


class BadRequestException(TechreadException):
    """ Raised when the request body cannot be interpreted by the server.
    This indicates that
    (i) someone has fiddled with with request body, or
    (ii) the server API version has been updated and the integration tests
        did not catch the problem.
    If you encounter this exception, please let us know
    """


class ResourceNotFoundException(TechreadException):
    """ Raised when we encounter a 404
    """


class UnsupportedMediaType(TechreadException):
    """ Raised when the uploaded file has a format
    that is not supported by the api
    """


class LicenseError(Exception):
    """ Error raised when the license information is
    incorrect
    """



from decimal import Decimal
from typing import Optional

from .feature import W24FeatureModel
from .gender import W24Gender
from .size import W24SizeTolerance

class W24ToleranceFeature(W24FeatureModel):
    """Characterization of a Tolerance Feature.

    Attributes:
        gender: Gender (male or female) of the tolerance feature.
            This is determined by checking whether the tolerance feature is
            located on the outer contour of the part or inside the part.
            When the outer contour is unavailable (e.g., in detail drawings),
            the gender is set to None.

        length: Length of the slug corresponding to the tolerance feature.

    """
    gender: Optional[W24Gender]

    length: Optional[Decimal]

    tolerance: W24SizeTolerance


import re
from os import path

from setuptools import setup

VERSIONFILE = "werk24/_version.py"
NAME = "werk24"


def _get_version(version_file: str) -> str:
    version_pattern = r"^__version__ = ['\"]([^'\"]*)['\"]"
    with open(version_file, "rt") as file_handle:
        match = re.search(version_pattern, file_handle.read(), re.M)
    if match:
        return match.group(1)

    raise RuntimeError(
        "Unable to find version string in %s." %
        (VERSIONFILE,))


def _get_description() -> str:
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md')) as file_handle:
        return file_handle.read()


setup(
    name=NAME,
    version=_get_version(VERSIONFILE),
    author="W24 Service GmbH",
    author_email="info@werk24.io",
    description="Werk24 Client to read PDF- and Image-based "
    "Technical Drawings / Engineering Drawings",
    url="https://werk24.io",
    entry_points={
        "console_scripts": [
            "w24cli=werk24.cli.w24cli:main",
            "w24gui=werk24.gui.w24gui:main",
        ]
    },
    extras_require={
        "gui": ["PyQt5", "pillow"]
    },
    license='commercial',
    packages=[
        "werk24",
        "werk24.cli",
        "werk24.gui",
        "werk24.models",
    ],
    include_package_data=True,

    package_data={"werk24": ["assets/*"]},
    install_requires=[
        "aiohttp >= 3.8.3",
        "boto3 >= 1.14.44",
        "devtools >=0.9.0",
        "pydantic >= 1.4",
        "python-dotenv>=0.10.1",
        "websockets >= 10.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Image Recognition"],
    keywords=[
        'Digitisation', 'Digitization', 'Engineering Drawing',
        'Engineering Drawings', 'Technical Drawing', 'Technical Drawings',
        'CAD', 'CAD Drawing', 'Data Extraction', 'Information Extraction',
        'Model Based Definition', 'EN10027', 'ISO2768', 'Title Block',
        'General Tolerances', 'Material', 'Drawing ID', 'Drawing Designation',
        'Product Manufacturing Information', 'PMI', 'Scanned Document',
        'Bill of Material', 'BOM', 'Anonymiziation', 'RFQ', 'GD&T',
        'General Dimensioning and Toleration', 'Vectorization'
    ],
    python_requires='>=3.7.1',
    project_urls={
        "Documentation": "https://docs.werk24.io/"
    },
    long_description_content_type='text/markdown',
    long_description=_get_description(),
)



""" Module handling the authentication
"""
import base64
import hashlib
import hmac
import time
from typing import Optional, Tuple

import boto3
from botocore.exceptions import ClientError

from werk24.exceptions import UnauthorizedException


class AuthClient:
    """ Client Module that handles the authentication
    with AWS Cognito.

    Raises:
        UnauthorizedException: Raised when the user credentials are not
            accepted by AWS Cognito

        RuntimeError: Raised when the server behaves in a very unexpected
            way; e.g., when AWS changed the protocol
    """

    def __init__(
        self,
        cognito_region: str,
        cognito_identity_pool_id: str,
        cognito_user_pool_id: str,
        cognito_client_id: str,
        cognito_client_secret: str
    ):

        # store the settings
        self._cognito_region = cognito_region
        self._cognito_identity_pool_id = cognito_identity_pool_id
        self._cognito_user_pool_id = cognito_user_pool_id
        self._cognito_client_id = cognito_client_id
        self._cognito_client_secret = cognito_client_secret

        # make empty references to the username and password
        self.username: Optional[str] = None
        self._password: Optional[str] = None

        # make an empty reference ot the jwt_token
        self.token: Optional[str] = None
        self.expires_at: Optional[float] = None
        self.refresh_token: Optional[str] = None

    def register(self, username: str, password: str) -> None:
        """ Store the username and password locally so
        that it can be used to obtain the token

        Arguments:
            username {str} -- Username
            password {str} -- Password
        """
        self.username = username
        self._password = password

    def _get_generic_identity(self) -> Tuple[str, str]:
        """ The AWS Cognito User Pools can only be accessed with
        credentials (even if they are generic). This function
        calls the AWS Cognito IDENTITY POOL to obtain the generic
        and unpriviledged credientials

        Raises:
            UnauthorizedException: Raise when we are not able to
                obtain unpriviledged credentials

        Returns:
            Tuple[str, str] -- Access Key, Secret Key Tuple
        """

        # make the identity client
        try:
            identity_client = boto3.client(
                'cognito-identity',
                self._cognito_region)

            # get a new identity id
            identity_response = identity_client.get_id(
                IdentityPoolId=self._cognito_identity_pool_id)
            identity_id = identity_response['IdentityId']

            # obtain the associated credentials
            credentials_response = identity_client \
                .get_credentials_for_identity(IdentityId=identity_id)
            credentials = credentials_response['Credentials']

        except KeyError:
            raise UnauthorizedException("Invalid Cognito configuration")

        except ClientError:  # pylint: disable=try-except-raise
            raise

        # get the access key / secret key
        access_key = credentials.get('AccessKeyId')
        secret_key = credentials.get('SecretKey')
        if access_key is None or secret_key is None:
            raise UnauthorizedException(
                "Unable to obtain Access and Secret Key from "
                "Cognito Identity Pool")

        # that's it
        return access_key, secret_key

    def _make_cognito_client(self) -> boto3.session.Session.client:
        """ Make the Cognito Client to communicate with
        AWS Cognito

        Returns:
            boto3.session.Session.client -- Boto3 Client
        """

        try:
            # before we can aws cognito USER POOL client, we need
            # to obtain a generic identity from the IDENTIY POOL
            access_key, secret_key = self._get_generic_identity()

            # with this information, we can now generate the client
            return boto3.client(
                "cognito-idp",
                region_name=self._cognito_region,
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key)
        except ClientError:
            raise UnauthorizedException("Cognito IDP Client Error")

    def _make_cognito_secret_hash(self, username: str) -> str:
        """ Make the keyed-hash message authentication code (HMAC) calculated
        using the secret key of a user pool client and username plus the client
        ID in the message.

        Arguments:
            username {str} -- Username

        Returns:
            str -- Cognito Secret Hash
        """

        # make the message
        message = username + self._cognito_client_id

        # make the secret
        dig = hmac.new(
            self._cognito_client_secret.encode("UTF-8"),
            msg=message.encode('UTF-8'),
            digestmod=hashlib.sha256).digest()

        # turn the secret into a str object
        return base64.b64encode(dig).decode()

    def login(
        self
    ) -> None:
        """ Login with AWS Cognito

        Raises:
            UnauthorizedException: Raised when the user credentials
                were not accepted by Cognito
        """
        if self.token is None or self._token_has_expired():
            self._login()

    def _token_has_expired(
        self,
        slack_minutes: int = 60
    ) -> bool:
        """ Check whether the token has already expired

        Returns:
            bool: True if exptired, False otherwise.
        """
        if self.expires_at is None:
            return True

        return time.time() > self.expires_at - slack_minutes

    def _login(self):
        """ Preform the login with AWS Cognito

        Raises:
            UnauthorizedException: Raised when the user is not authorized
                or something went wrong during the authentication process.
        """

        # there is no point in trying to log in if there is no
        # username or password
        if self.username is None or self._password is None:
            raise UnauthorizedException("No username / password provided")

        # make the connection to aws
        cognito_client = self._make_cognito_client()

        # make the authentication data
        auth_data = {
            'USERNAME': self.username,
            'PASSWORD': self._password,
            'SECRET_HASH': self._make_cognito_secret_hash(self.username)
        }

        # get the jwt token from AWS cognito
        try:
            resp = cognito_client.initiate_auth(
                AuthFlow='USER_PASSWORD_AUTH',
                AuthParameters=auth_data,
                ClientId=self._cognito_client_id)

        # We will receive an error message directly from AWS Cognito
        # if anything goes wrong. The error message will be valuable,
        # as it allows the user to differentiate between disabled
        # accounts and incorrect credentials
        except cognito_client.exceptions.NotAuthorizedException:
            raise UnauthorizedException()

        # store the jwt token
        try:
            result = resp['AuthenticationResult']
            self.token = result['IdToken']
            self.expires_at = time.time() + int(result['ExpiresIn'])
            self.refresh_token = result['RefreshToken']
        except KeyError:
            raise UnauthorizedException(
                "Unable to obtain JWT Token from AWS Cognito.")



from decimal import Decimal
from typing import Optional
from pydantic import BaseModel
from enum import Enum


class W24TestCriticality(str, Enum):
    CRITICAL_TO_DELIVERY = "CTD"
    CRITICAL_TO_FUNCTION = "CTF"
    CRITICAL_TO_PRICE = "CTP"
    CRITICAL_TO_QUALITY = "CTQ"
    CRITICAL_TO_SAFETY = "CTS"


class W24TestDimension(BaseModel):
    """ Test Dimension attributes that can either be attached
    to a MeasureLabel or a GDTLabel

    Attributes:
        blurb: Blurb of the complete label for human consumption

        criticality: Some measures can be critical to Function,
            others critical to price. The TestDimensions sometimes
            highlight this with a shorthand (e.g., CTF)

        rate: Test rate as percentage. If a 100% inspection rate has been
            defined, rate will be Decimal('100'). If not defined, the
            rate will be None.
    """
    blurb: str
    criticality: Optional[W24TestCriticality] = None
    rate: Optional[Decimal] = None



from typing import List, Callable, Dict
from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QHeaderView
from PyQt5.QtCore import QRect
from werk24.models.gdt import W24GDT


class W24GuiGdtTable(QTableWidget):

    @classmethod
    def from_gdt_list(
            cls,
            measure_list: List[W24GDT]
    ) -> 'W24GuiGdtTable':

        # make a simple config file for the current table
        config: Dict[str, Callable] = {
            "Polygon": cls._format_gdt_polygon,
            "Blurb": cls._format_gdt_blurb,
        }

        # make the associated table
        table_widget = W24GuiGdtTable()
        table_widget.setColumnCount(len(config))
        table_widget.setRowCount(len(measure_list))
        table_widget.verticalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents)

        # make the table content
        table_widget.setHorizontalHeaderLabels(config.keys())
        for i, cur_measure in enumerate(measure_list):
            for j, callable_ in enumerate(config.values()):
                content = callable_(cur_measure)
                table_widget.setItem(i, j, QTableWidgetItem(content))
            table_widget.resizeColumnToContents(i)

        table_widget.setGeometry(0, 0, 200, 200)
        table_widget.resizeRowsToContents()
        return table_widget

    @staticmethod
    def _format_gdt_blurb(gdt: W24GDT) -> str:
        return f"{gdt.frame.blurb}"

    @staticmethod
    def _format_gdt_polygon(gdt: W24GDT) -> str:
        return " - ".join([
            f"{pt[0]:.4f} x {pt[1]:.4f}"
            for pt in gdt.bounding_polygon])



from pydantic import BaseModel, validator
from decimal import Decimal
from typing import Optional, List, Dict, Any, Union, Type

from .gender import W24Gender
from .thread import W24Thread, W24ThreadISOMetric, W24ThreadUTS, W24ThreadWhitworth



def deserialize_thread(
    raw: Union[Dict[str, Any], W24Thread],
) -> W24Thread:
    """ Deserialize a specific ask in its raw form

    Args:
        raw (Dict[str, Any]): Raw Ask as it arrives from the
            json deserializer

    Returns:
        W24AskUnion: Corresponding ask type
    """
    if isinstance(raw, dict):
        ask_type = _deserialize_thread_type(raw.get('thread_type', ''))
        return ask_type.parse_obj(raw)

    if isinstance(raw, W24Thread):
        return raw

    raise ValueError(f"Unsupported value type '{type(raw)}'")

def _deserialize_thread_type(
    ask_type: str
) -> Type[W24Thread]:
    """ Get the Ask Class from the ask type

    Args:
        ask_type (str): Ask type in question

    Raises:
        ValueError: Raised if ask type is unknown

    Returns:
        str: Name of the AskObject
    """
    class_ = {
        "ISO_METRIC": W24ThreadISOMetric,
        "WHITWORTH": W24ThreadWhitworth,
        "UTS": W24ThreadUTS,
    }.get(ask_type, None)

    if class_ is None:
        raise ValueError(f"Unknown Ask Type '{ask_type}'")

    return class_


class W24ThreadElement(BaseModel):
    """Characterization of a Thread Element

    Attributes:
        quantity (int): Number of equivalent elements that are
            indicated by one label (e.g., 3xM5).

        gender: Gender (male or female) of the thread feature.
            This is determined by checking whether the thread is
            located on the outer contour of the part or inside the part.
            When the outer contour is unavailable (e.g., in detail drawings),
            the gender is set to None.

        length: Length of the stud corresponding
            to the thread. Please keep in mind that a thus can
            contain multiple threads with different lengths. This
            is considered in the threads themselves.
            NOTE: we are using the word length here to differentiate
            it from the thread depth which describes the difference
            between the major and minor radii.

        threads (List[W24Thread]): List of Threads that are positioned
            on the ThreadElements. This is a list to support multi-threads

        NOTE: Tapers are currently not considered

        NOTE: Future implementations might also consider the inclination
        in two angles relative to the front view.
    """
    quantity: int

    gender: Optional[W24Gender]

    length: Optional[Decimal]

    threads: List[W24Thread]

    @validator('threads', pre=True)
    def ask_list_validator(
        cls,
        raw: List[Dict[str, Any]]
    ) -> List[W24Thread]:
        """ Validator to de-serialize the asks. The de-serialization
        is based on the ask_type attribute of the object. Pydantic
        does not support this out-of-the box

        Args:
            raw (Dict[str, Any]): Raw json of the asks list

        Returns:
            List[W24AskUnion]: List of deserialized Asks
        """
        return [deserialize_thread(t) for t in raw]
